use embassy_rp::pio::{Common, Config, Instance, PioPin, StateMachine};
use embassy_time::Timer;
use fixed::traits::ToFixed;
use pio_proc::pio_file;

use crate::DhtError;

pub struct DhtPio<'d, PIO: Instance, const SM: usize> {
    sm: StateMachine<'d, PIO, SM>,
}

impl<'d, PIO: Instance, const SM: usize> DhtPio<'d, PIO, SM> {
    pub fn new(
        pio: &mut Common<'d, PIO>,
        mut sm: StateMachine<'d, PIO, SM>,
        dht_pin: impl PioPin,
    ) -> Self {
        let program = pio_file!("./src/dht.pio");

        let pin = pio.make_pio_pin(dht_pin);

        let mut config = Config::default();
        config.use_program(&pio.load_program(&program.program), &[]);
        config.set_out_pins(&[&pin]);
        config.set_set_pins(&[&pin]);
        config.set_in_pins(&[&pin]);
        config.shift_out.threshold = 32;
        config.clock_divider = 125.to_fixed();

        sm.set_pin_dirs(embassy_rp::pio::Direction::Out, &[&pin]);
        sm.set_config(&config);
        sm.set_enable(true);

        Self { sm }
    }

    pub async fn read_data(&mut self) -> Result<(u16, u16), DhtError> {
        let mut timeout = 2000;
        let mut raw: [Option<u32>; 2] = [None; 2];

        self.sm.tx().wait_push(1).await;

        while timeout > 0 && raw[1].is_none() {
            match self.sm.rx().try_pull() {
                Some(d) => {
                    if raw[0].is_none() {
                        raw[0] = Some(d);
                    } else {
                        raw[1] = Some(d);
                    }
                }
                None => (),
            }

            Timer::after_millis(1).await;
            timeout -= 1;
        }

        if timeout <= 0 {
            self.sm.restart();
            return Err(DhtError::Timeout);
        }

        let data = raw[0].unwrap();
        let crc = raw[1].unwrap();

        if Self::compute_crc(data) != crc {
            return Err(DhtError::CrcMismatch(
                raw[0].unwrap_or(0),
                raw[1].unwrap_or(0),
            ));
        }

        Ok((
            (data & 0x0000FFFF) as u16,
            ((data & 0xFFFF0000) >> 16) as u16,
        ))
    }

    fn compute_crc(data: u32) -> u32 {
        let mut crc: u32 = 0;
        crc += data & 0x000000FF;
        crc += (data & 0x0000FF00) >> 8;
        crc += (data & 0x00FF0000) >> 16;
        crc += (data & 0xFF000000) >> 24;

        crc % 256
    }
}

impl<'d, PIO: Instance, const SM: usize> Drop for DhtPio<'d, PIO, SM> {
    fn drop(&mut self) {
        self.sm.set_enable(false);
    }
}
